'use strict';

Object.defineProperty(exports, '__esModule', { value: true });

var core = require('@babel/core');
var path = require('path');

const cache = {};

const shouldUsePublicPath = (state, parent) => {
  var _parent$source, _parent$source$value;

  if (!state.publicPath) {
    return false;
  }

  const src = parent == null ? void 0 : (_parent$source = parent.source) == null ? void 0 : (_parent$source$value = _parent$source.value) == null ? void 0 : _parent$source$value[0];
  return src === '.' || src === '/' || src == null;
};

const typeSourceVisitor = {
  Class(path, state) {
    var _path$node$id;

    if (state.name) {
      return;
    }

    const name = (_path$node$id = path.node.id) == null ? void 0 : _path$node$id.name;

    if (name !== state.typeName) {
      return;
    }

    const key = `${state.filename}/${state.typeName}`;

    if (!cache[key]) {
      const usePublicPath = shouldUsePublicPath(state, null);
      const value = `${usePublicPath ? state.publicPath : state.filename}/${state.typeName}`;
      cache[key] = value;
    }

    state.name = cache[key];
  },

  TSType(path, state) {
    var _parent$id;

    if (state.name) {
      return;
    }

    const parent = path.parent;
    const parentName = parent == null ? void 0 : (_parent$id = parent.id) == null ? void 0 : _parent$id.name;

    if (parentName !== state.typeName) {
      return;
    }

    const key = `${state.filename}/${state.typeName}`;

    if (!cache[key]) {
      const usePublicPath = shouldUsePublicPath(state, parent);
      const value = `${usePublicPath ? state.publicPath : state.filename}/${state.typeName}`;
      cache[key] = value;
    }

    state.name = cache[key];
  },

  TSInterfaceDeclaration(path, state) {
    if (state.name) {
      return;
    }

    if (path.node.id.name !== state.typeName) {
      return;
    } // FIXME: not sure what type of node parent is


    const parent = path.parent;
    const key = `${state.filename}/${state.typeName}`;

    if (!cache[key]) {
      const usePublicPath = shouldUsePublicPath(state, parent);
      const value = `${usePublicPath ? state.publicPath : state.filename}/${state.typeName}`;
      cache[key] = value;
    }

    state.name = cache[key];
  },

  ImportSpecifier(path$1, state) {
    if (state.name) {
      return;
    }

    if (path$1.node.local.name !== state.typeName) {
      return;
    }

    const parent = path$1.parent;
    let source = parent.source.value;
    Object.keys(state.pathAlias).forEach(key => {
      if (source.startsWith(key)) {
        source = source.replace(key, state.pathAlias[key]);
      }
    });

    if (source.charAt(0) === '.') {
      source = path.join(path.dirname(state.filename), source);
    }

    const key = `${source}/${path$1.node.imported.name}`;

    if (!cache[key]) {
      const usePublicPath = shouldUsePublicPath(state, parent);
      const value = usePublicPath ? `${state.publicPath}/${path$1.node.imported.name}` : key;
      cache[key] = value;
    }

    state.name = cache[key];
  }

}; // given a ts type, we try to find the "actual" name of it, including an appropriate prefix

function getConcreteTypeName(typeNode, filename, publicPath, pathAlias, programPath) {
  // for a type reference we need to find where it is declared first
  if (core.types.isTSTypeReference(typeNode) || core.types.isTSExpressionWithTypeArguments(typeNode)) {
    var _typeNode$typeName$na, _typeNode$typeName, _typeNode$expression;

    // @ts-ignore
    const name = (_typeNode$typeName$na = (_typeNode$typeName = typeNode.typeName) == null ? void 0 : _typeNode$typeName.name) != null ? _typeNode$typeName$na : (_typeNode$expression = typeNode.expression) == null ? void 0 : _typeNode$expression.name;

    if (name == null) {
      return null;
    } // if we're dealing with a node module or global type, we can resolve it right here


    if (name === 'NodeModule' || name === 'Global') {
      var _typeNode$typeParamet, _typeNode$typeParamet2, _param$literal;

      const param = (_typeNode$typeParamet = typeNode.typeParameters) == null ? void 0 : (_typeNode$typeParamet2 = _typeNode$typeParamet.params) == null ? void 0 : _typeNode$typeParamet2[0];
      const value = param == null ? void 0 : (_param$literal = param.literal) == null ? void 0 : _param$literal.value;

      if (value) {
        if (name === 'Global') {
          return `type:global:${value}`;
        }

        return value;
      }
    } // we need to traverse the file to find the type declaration


    const state = {
      filename,
      publicPath,
      pathAlias,
      typeName: name,
      name: null
    };
    programPath.traverse(typeSourceVisitor, state);

    if (state.name) {
      return `type:${state.name}`;
    } // we have an actual type but we can't find it in the file
    // presumably it's a global type...


    return `type:global:${name}`;
  } // classes are like magic snowflakes that can be used as both concretes and types


  if (core.types.isClass(typeNode) || core.types.isClassExpression(typeNode)) {
    const name = typeNode.id.name;
    const key = `${filename}/${name}`;

    if (!cache[key]) {
      const usePublicPath = shouldUsePublicPath({
        publicPath,
        pathAlias,
        filename: null,
        typeName: null,
        name: null
      }, null);
      const value = `${usePublicPath ? publicPath : filename}/${name}`;
      cache[key] = value;
    }

    return `type:${cache[key]}`;
  }

  if (core.types.isTSTypeLiteral(typeNode) || core.types.isTSFunctionType(typeNode)) {
    throw new Error('Currently registering with a literal type is not supported');
  }
}

const tsTypeAnnotationVisitor = {
  TSTypeAnnotation(path, state) {
    const name = getConcreteTypeName(path.node.typeAnnotation, state.filename, state.publicPath, state.pathAlias, state.programPath);
    state.key = name == null ? 'unknown' : name;
  }

};

const getFunctionParams = (path, deps, filename, publicPath, pathAlias, programPath) => {
  [].concat(path.get('params')).forEach(path => {
    const ctx = {
      key: path.node.name,
      filename,
      programPath,
      publicPath,
      pathAlias
    };
    path.traverse(tsTypeAnnotationVisitor, ctx);
    deps.push(ctx.key);
  });
};

const classConstructorVisitor = {
  ClassMethod(path, state) {
    const {
      deps,
      filename,
      programPath,
      publicPath,
      pathAlias
    } = state; // @ts-ignore

    if (path.node.key.name === 'constructor') {
      getFunctionParams(path, deps, filename, publicPath, pathAlias, programPath);
    }
  }

};
const linkedVariableVisitor = {
  Class(path, state) {
    var _path$node$id;

    const {
      deps,
      name,
      filename,
      programPath,
      publicPath,
      pathAlias
    } = state;

    if (((_path$node$id = path.node.id) == null ? void 0 : _path$node$id.name) === name) {
      path.traverse(classConstructorVisitor, {
        deps,
        filename,
        programPath,
        publicPath,
        pathAlias
      });
    }
  },

  ArrowFunctionExpression(path, state) {
    var _parent$id;

    const {
      deps,
      name,
      filename,
      programPath,
      publicPath,
      pathAlias
    } = state;
    const {
      parent
    } = path; // @ts-ignore

    if ((parent == null ? void 0 : (_parent$id = parent.id) == null ? void 0 : _parent$id.name) === name) {
      getFunctionParams(path, deps, filename, publicPath, pathAlias, programPath);
    }
  },

  FunctionDeclaration(path, state) {
    const {
      deps,
      name,
      filename,
      programPath,
      publicPath,
      pathAlias
    } = state;
    const {
      node
    } = path;

    if (node && node.id && node.id.name === name) {
      getFunctionParams(path, deps, filename, publicPath, pathAlias, programPath);
    }
  },

  FunctionExpression(path, state) {
    var _parent$id2;

    const {
      deps,
      name,
      filename,
      programPath,
      publicPath,
      pathAlias
    } = state;
    const {
      parent
    } = path; // @ts-ignore

    if ((parent == null ? void 0 : (_parent$id2 = parent.id) == null ? void 0 : _parent$id2.name) === name) {
      getFunctionParams(path, deps, filename, publicPath, pathAlias, programPath);
    }
  }

};
function extractFunctionParameterTypes(programPath, arg, filename, publicPath, pathAlias) {
  const deps = [];
  const state = {
    deps,
    programPath,
    filename,
    publicPath,
    pathAlias
  };

  if (core.types.isIdentifier(arg)) {
    programPath.traverse(linkedVariableVisitor, Object.assign({
      name: arg.node.name
    }, state));
  } else if (core.types.isClass(arg)) {
    arg.traverse(classConstructorVisitor, state);
  } else if (core.types.isArrowFunctionExpression(arg)) {
    getFunctionParams(arg, deps, filename, publicPath, pathAlias, programPath);
  } else if (core.types.isFunctionExpression(arg)) {
    getFunctionParams(arg, deps, filename, publicPath, pathAlias, programPath);
  }

  return deps;
}

const visitor = {
  Class(path, state) {
    var _path$node$id;

    if (((_path$node$id = path.node.id) == null ? void 0 : _path$node$id.name) === state.typeName) {
      var _path$node$implements;

      const aliases = (_path$node$implements = path.node.implements) == null ? void 0 : _path$node$implements.map(node => {
        return getConcreteTypeName(node, state.filename, state.publicPath, state.pathAlias, state.programPath);
      }).filter(Boolean);

      if (aliases == null ? void 0 : aliases.length) {
        state.aliases = aliases;
      }
    }
  }

};
function getImplements(node, filename, publicPath, pathAlias, programPath) {
  if (core.types.isIdentifier(node)) {
    const state = {
      aliases: null,
      typeName: node.name,
      filename,
      publicPath,
      pathAlias,
      programPath
    };
    programPath.traverse(visitor, state);
    return state.aliases;
  }

  if (core.types.isClassExpression(node)) {
    var _node$implements;

    const aliases = (_node$implements = node.implements) == null ? void 0 : _node$implements.map(node => {
      return getConcreteTypeName(node, filename, publicPath, pathAlias, programPath);
    }).filter(Boolean);
    return (aliases == null ? void 0 : aliases.length) ? aliases : null;
  }
}

const visitor$1 = {
  Class(path, state) {
    var _path$node$id;

    if (((_path$node$id = path.node.id) == null ? void 0 : _path$node$id.name) === state.typeName) {
      state.name = getConcreteTypeName(path.node, state.filename, state.publicPath, state.pathAlias, state.programPath);
    }
  }

};
function getTypeofNode(node, {
  filename,
  publicPath,
  pathAlias
}, programPath) {
  if (core.types.isIdentifier(node)) {
    const state = {
      name: null,
      typeName: node.name,
      filename,
      publicPath,
      pathAlias,
      programPath
    };
    programPath.traverse(visitor$1, state);
    return state.name;
  }

  if (core.types.isClassExpression(node)) {
    return getConcreteTypeName(node, filename, publicPath, pathAlias, programPath);
  }
}

function getTypeParameter(path, i = 0) {
  var _path$node, _path$node$typeParame, _path$node$typeParame2;

  return (_path$node = path.node) == null ? void 0 : (_path$node$typeParame = _path$node.typeParameters) == null ? void 0 : (_path$node$typeParame2 = _path$node$typeParame.params) == null ? void 0 : _path$node$typeParame2[i];
}

function isJpexCall(path, identifier, key) {
  const callee = path.node.callee; // must be a member expression i.e. jpex.whatever()

  if (!core.types.isMemberExpression(callee)) {
    return false;
  } // must be a valid identifier i.e. jpex
  // @ts-ignore


  if (!identifier.includes(callee.object.name)) {
    return false;
  }

  if (Array.isArray(key)) {
    // @ts-ignore
    return key.includes(callee.property.name);
  } // @ts-ignore


  return callee.property.name === key;
}

exports.extractFunctionParameterTypes = extractFunctionParameterTypes;
exports.getConcreteTypeName = getConcreteTypeName;
exports.getImplements = getImplements;
exports.getTypeParameter = getTypeParameter;
exports.getTypeofNode = getTypeofNode;
exports.isJpexCall = isJpexCall;
